# -*- coding: utf-8 -*-

"""

dryjq.__main__

dryjq (or python3 -m dryjq) command line script

Copyright (C) 2022 Rainer Schwarzbach

This file is part of dryjq.

dryjq is free software: you can redistribute it and/or modify
it under the terms of the MIT License.

dryjq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the LICENSE file for more details.

"""


import logging
import sys

from typing import List, Optional

from dryjq import commandline
from dryjq import commons
from dryjq import queries


#
# Functions
#


def main(args: Optional[List[str]] = None) -> int:
    """Use the query parser to construct
    an access.Path subclass instance from the original query.
    Then, execute the program using that instance together with
    the provided command line arguments.

    :param args: the list of command line arguments,
        or None to go with the default (sys.argv).
    :returns: the script returncode
    """
    program = commandline.Program(args)
    try:
        parsed_query = queries.Parser().parse_query(
            program.arguments.query,
            separator_codepoint=ord(program.arguments.separator),
            subscript_indicators_pair=commons.CharactersPair(
                program.arguments.subscript_indicators
            ),
        )
    except (
        queries.MalformedQueryError,
        queries.IllegalStateError,
        ValueError,
        TypeError,
    ) as error:
        logging.error(error)
        return commandline.RETURNCODE_ERROR
    #
    return program.execute(parsed_query)


if __name__ == "__main__":
    sys.exit(main())  # NOT TESTABLE


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-

"""

test_commons

Unit test the commons module

Copyright (C) 2022 Rainer Schwarzbach

This file is part of dryjq.

dryjq is free software: you can redistribute it and/or modify
it under the terms of the MIT License.

dryjq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the LICENSE file for more details.

"""


from unittest import TestCase

from dryjq import commons


class CharactersPair(TestCase):

    """CharactersPair tests"""

    def test_properties(self):
        """Test properties"""
        pair = commons.CharactersPair("()")
        for tested_property, expected_value in (
            ("first", ord("(")),
            ("last", ord(")")),
            ("open", ord("(")),
            ("close", ord(")")),
            ("source", "()"),
            ("both", (ord("("), ord(")"))),
        ):
            with self.subTest(
                tested_property=tested_property, expected_value=expected_value
            ):
                self.assertEqual(
                    getattr(pair, tested_property),
                    expected_value,
                )
            #
        #

    def test_single_character(self):
        """Test initialization with a single character"""
        pair = commons.CharactersPair("/")
        codepoint = ord("/")
        for tested_property in ("first", "last", "open", "close"):
            with self.subTest(
                tested_property=tested_property, codepoint=codepoint
            ):
                self.assertEqual(
                    getattr(pair, tested_property),
                    codepoint,
                )
            #
        #

    def test_exceptions(self):
        """Test initialization errors"""
        for parameter, expected_exception in (
            ("too long", ValueError),
            ("", IndexError),
        ):
            with self.subTest(parameter=parameter):
                self.assertRaises(
                    expected_exception,
                    commons.CharactersPair,
                    parameter,
                )
            #
        #

    def test_comparisons(self):
        """Test comparisons and hash equality"""
        for source1, source2, equal, hash_equal in (
            ("[]", "()", False, False),
            ("//", "/", True, False),
            ("!", "!", True, True),
        ):
            with self.subTest(source1=source1, source2=source2):
                cp1 = commons.CharactersPair(source1)
                cp2 = commons.CharactersPair(source2)
                if equal:
                    self.assertEqual(cp1, cp2)
                else:
                    self.assertNotEqual(cp1, cp2)
                #
                if hash_equal:
                    self.assertEqual(hash(cp1), hash(cp2))
                else:
                    self.assertNotEqual(hash(cp1), hash(cp2))
                #
            #
        #


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-

"""

test_main_file

Unit test the command line interface
with input from a file

Copyright (C) 2022 Rainer Schwarzbach

This file is part of dryjq.

dryjq is free software: you can redistribute it and/or modify
it under the terms of the MIT License.

dryjq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the LICENSE file for more details.

"""


import io
import os
import secrets
import tempfile
import time

from unittest import TestCase
from unittest.mock import patch

from dryjq import __main__

from . import data
from .commons import GenericCallResult


class CommandLineResult(GenericCallResult):

    """Command line call result"""

    @classmethod
    def do_call(cls, *args, **kwargs):
        """Do the real function call"""
        del kwargs
        return __main__.main(list(args))


class TestSimple(TestCase):

    """Simple test module using patched stdin and stdout"""

    # tempdir = None

    def setUp(self):
        """Create a temporary directory"""
        # pylint: disable=consider-using-with
        self.tempdir = tempfile.TemporaryDirectory()
        # pylint: enable=consider-using-with
        self.testfile_name = os.path.join(
            self.tempdir.name, f"{secrets.token_urlsafe(6)}.yaml"
        )

    def tearDown(self):
        """Cleanup the temporary directory"""
        self.tempdir.cleanup()

    def __create_file(self, content=None):
        """Create a file in the temporary directory"""
        if content:
            with open(
                self.testfile_name,
                mode="w",
                encoding="utf-8",
            ) as testdata_file:
                testdata_file.write(content)
            #
        #

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_get_single_value(self, mock_stdout):
        """Get a single value"""
        self.__create_file(content=data.INPUT_YAML_ANIMALS)
        result = CommandLineResult.from_call(
            ".felines.cats.big[1]",
            self.testfile_name,
            stdout=mock_stdout,
        )
        self.assertEqual(result.stdout, data.EXPECT_CATS_BIG_1)
        self.assertEqual(result.returncode, data.RETURNCODE_OK)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_get_subtree(self, mock_stdout):
        """Get a subtree"""
        self.__create_file(content=data.INPUT_YAML_ANIMALS)
        result = CommandLineResult.from_call(
            ".canines",
            self.testfile_name,
            stdout=mock_stdout,
        )
        self.assertEqual(result.stdout, data.EXPECT_YAML_CANINES)
        self.assertEqual(result.returncode, data.RETURNCODE_OK)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_missing_list_value(self, mock_stdout):
        """Try to get a missing list value"""
        self.__create_file(content=data.INPUT_YAML_ANIMALS)
        missing_index = 5
        with self.assertLogs(None, level="ERROR") as log_cm:
            result = CommandLineResult.from_call(
                f".felines.cats.big[{missing_index}]",
                self.testfile_name,
                stdout=mock_stdout,
            )
        #
        self.assertIn(
            f"TraversalPath('felines', 'cats', 'big', {missing_index!r})"
            " not found!",
            log_cm.output[0],
        )
        self.assertEqual(result.stdout, "")
        self.assertEqual(result.returncode, data.RETURNCODE_ERROR)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_missing_mapping_value(self, mock_stdout):
        """Try to get a missing mapping value"""
        self.__create_file(content=data.INPUT_YAML_ANIMALS)
        missing_key = "rodents"
        with self.assertLogs(None, level="ERROR") as log_cm:
            result = CommandLineResult.from_call(
                f".{missing_key}",
                self.testfile_name,
                stdout=mock_stdout,
            )
        #
        self.assertIn(
            f"TraversalPath({missing_key!r}) not found!",
            log_cm.output[0],
        )
        self.assertEqual(result.stdout, "")
        self.assertEqual(result.returncode, data.RETURNCODE_ERROR)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_indexing_scalar_value(self, mock_stdout):
        """Try to get an indexed value from a scalar"""
        self.__create_file(content=data.INPUT_YAML_ANIMALS)
        wrong_index = 1
        path_to_value = ".felines.cats.big[0]"
        mock_stdout.seek(0)
        mock_stdout.truncate()
        with self.assertLogs(None, level="ERROR") as log_cm:
            result = CommandLineResult.from_call(
                f"{path_to_value}[{wrong_index}]",
                self.testfile_name,
                stdout=mock_stdout,
            )
        #
        self.assertIn(
            "Cannot traverse through a leaf",
            log_cm.output[0],
        )
        self.assertEqual(result.stdout, "")
        self.assertEqual(result.returncode, data.RETURNCODE_ERROR)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_key_indexing_list(self, mock_stdout):
        """Try to get a value indexed by a key from a list"""
        self.__create_file(content=data.INPUT_YAML_ANIMALS)
        wrong_index = "lion"
        with self.assertLogs(None, level="ERROR") as log_cm:
            result = CommandLineResult.from_call(
                f".felines.cats.big[{wrong_index}]",
                self.testfile_name,
                stdout=mock_stdout,
            )
        #
        self.assertIn(
            "list indices must be integers or slices, not str",
            log_cm.output[0],
        )
        self.assertEqual(result.stdout, "")
        self.assertEqual(result.returncode, data.RETURNCODE_ERROR)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_replace_single_value(self, mock_stdout):
        """Replace a single value"""
        self.__create_file(content=data.INPUT_YAML_ANIMALS)
        result = CommandLineResult.from_call(
            ".felines.cats.big[2] = panther",
            self.testfile_name,
            stdout=mock_stdout,
        )
        self.assertEqual(result.stdout, data.EXPECT_YAML_JAGUAR2PANTHER)
        self.assertEqual(result.returncode, data.RETURNCODE_OK)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_replace_subtree(self, mock_stdout):
        """Replace a subtree"""
        self.__create_file(content=data.INPUT_YAML_ANIMALS)
        result = CommandLineResult.from_call(
            f".felines.cats = {data.REPLACEMENT_YAML_CATS_TAXONOMY}",
            self.testfile_name,
            stdout=mock_stdout,
        )
        self.assertEqual(result.stdout, data.EXPECT_YAML_CATS_TAXONOMY)
        self.assertEqual(result.returncode, data.RETURNCODE_OK)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_replace_subtree_inplace(self, mock_stdout):
        """Replace a subtree and modify the file in place"""
        self.__create_file(content=data.INPUT_YAML_ANIMALS)
        time.sleep(0.1)
        creation_time = os.stat(self.testfile_name).st_mtime
        result = CommandLineResult.from_call(
            "--inplace",
            f".felines.cats = {data.REPLACEMENT_YAML_CATS_TAXONOMY}",
            self.testfile_name,
            stdout=mock_stdout,
        )
        self.assertEqual(result.stdout, "")
        self.assertEqual(result.returncode, data.RETURNCODE_OK)
        self.assertGreater(os.stat(self.testfile_name).st_mtime, creation_time)
        with open(
            self.testfile_name, mode="r", encoding="utf-8"
        ) as changed_file:
            changed_data = changed_file.read()
        #
        self.assertEqual(changed_data, data.EXPECT_YAML_CATS_TAXONOMY)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_unchanged_file(self, mock_stdout):
        """Do a modification not changing original data"""
        self.__create_file(content=data.INPUT_YAML_ANIMALS)
        time.sleep(0.1)
        creation_time = os.stat(self.testfile_name).st_mtime
        with self.assertLogs(None, level="WARNING") as log_cm:
            result = CommandLineResult.from_call(
                "--inplace",
                ".felines.cats.big[2] = jaguar",
                self.testfile_name,
                stdout=mock_stdout,
            )
        #
        self.assertIn(
            "contents did not change",
            log_cm.output[0],
        )
        self.assertEqual(result.stdout, "")
        self.assertEqual(result.returncode, data.RETURNCODE_OK)
        self.assertEqual(os.stat(self.testfile_name).st_mtime, creation_time)
        with open(
            self.testfile_name, mode="r", encoding="utf-8"
        ) as unchanged_file:
            unchanged_data = unchanged_file.read()
        #
        self.assertEqual(unchanged_data, data.INPUT_YAML_ANIMALS)


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-

"""

test_queries

Unit test the queries module

Copyright (C) 2022 Rainer Schwarzbach

This file is part of dryjq.

dryjq is free software: you can redistribute it and/or modify
it under the terms of the MIT License.

dryjq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the LICENSE file for more details.

"""


from unittest import TestCase

from unittest.mock import patch

from serializable_trees.basics import ListNode, MapNode
from serializable_trees import TraversalPath, Tree

from dryjq import commons
from dryjq import queries


class Itemizer(TestCase):

    """Itemizer class test"""

    def setUp(self):
        """Initialize the itemizer"""
        self.itemizer = queries.Itemizer()

    def test_queries(self):
        """Test queries"""
        for query, expected_items in (
            (
                "",
                [
                    queries.QueryStartItem(),
                    queries.QueryEndItem(),
                ],
            ),
            (
                ".",
                [
                    queries.QueryStartItem(),
                    queries.SeparatorItem("."),
                    queries.QueryEndItem(),
                ],
            ),
            (
                " .",
                [
                    queries.QueryStartItem(),
                    queries.WhitespaceItem(" "),
                    queries.SeparatorItem("."),
                    queries.QueryEndItem(),
                ],
            ),
            (
                ".abc",
                [
                    queries.QueryStartItem(),
                    queries.SeparatorItem("."),
                    queries.LiteralItem("abc"),
                    queries.QueryEndItem(),
                ],
            ),
            (
                ".abc['yes'].xyz = [1, 2, 4]",
                [
                    queries.QueryStartItem(),
                    queries.SeparatorItem("."),
                    queries.LiteralItem("abc"),
                    queries.SubscriptOpenerItem("["),
                    queries.LiteralItem("'yes'"),
                    queries.SubscriptCloserItem("]"),
                    queries.SeparatorItem("."),
                    queries.LiteralItem("xyz"),
                    queries.WhitespaceItem(" "),
                    queries.AssignmentItem("="),
                    queries.WhitespaceItem(" "),
                    queries.SubscriptOpenerItem("["),
                    queries.LiteralItem("1,"),
                    queries.WhitespaceItem(" "),
                    queries.LiteralItem("2,"),
                    queries.WhitespaceItem(" "),
                    queries.LiteralItem("4"),
                    queries.SubscriptCloserItem("]"),
                    queries.QueryEndItem(),
                ],
            ),
        ):
            with self.subTest(query=query):
                self.assertEqual(
                    list(self.itemizer.itemize(query)),
                    expected_items,
                )
            #
        #

    def test_specialchars_variations(self):
        """Test variations of separator and subscript (...)
        ... indicator characters in queries
        """
        for query, separator, subscript_pair, expected_items in (
            (
                "/",
                "/",
                "[]",
                [
                    queries.QueryStartItem(),
                    queries.SeparatorItem("/"),
                    queries.QueryEndItem(),
                ],
            ),
            (
                "?abc",
                "?",
                "<>",
                [
                    queries.QueryStartItem(),
                    queries.SeparatorItem("?"),
                    queries.LiteralItem("abc"),
                    queries.QueryEndItem(),
                ],
            ),
            (
                "!abc('yes')!xyz = [1, 2, 4]",
                "!",
                "()",
                [
                    queries.QueryStartItem(),
                    queries.SeparatorItem("!"),
                    queries.LiteralItem("abc"),
                    queries.SubscriptOpenerItem("("),
                    queries.LiteralItem("'yes'"),
                    queries.SubscriptCloserItem(")"),
                    queries.SeparatorItem("!"),
                    queries.LiteralItem("xyz"),
                    queries.WhitespaceItem(" "),
                    queries.AssignmentItem("="),
                    queries.WhitespaceItem(" "),
                    queries.LiteralItem("[1,"),
                    queries.WhitespaceItem(" "),
                    queries.LiteralItem("2,"),
                    queries.WhitespaceItem(" "),
                    queries.LiteralItem("4]"),
                    queries.QueryEndItem(),
                ],
            ),
        ):
            with self.subTest(query=query):
                itemizer = queries.Itemizer(
                    separator_codepoint=ord(separator),
                    subscript_indicators_pair=commons.CharactersPair(
                        subscript_pair
                    ),
                )
                self.assertEqual(
                    list(itemizer.itemize(query)),
                    expected_items,
                )
            #
        #

    def test_errors(self):
        """Test itemizer errors"""
        for query, expected_exception, position in (
            (".abc[null", queries.UnclosedSubscriptError, 5),
            (".abc['def", queries.UnclosedQuoteError, 6),
        ):
            with self.subTest(query=query):
                self.assertRaisesRegex(
                    expected_exception,
                    f".* character position #{position}",
                    list,
                    self.itemizer.itemize(query),
                )
            #

    def test_warning(self):
        """Test the warning about subscript opener in subscript"""

        with self.assertLogs(None, level="WARNING") as log_cm:
            result = list(self.itemizer.itemize(".abc[def[ghi]jkl]"))
        #
        self.assertIn(
            "Possible error: Found subscript opener",
            log_cm.output[0],
        )
        self.assertEqual(
            result,
            [
                queries.QueryStartItem(),
                queries.SeparatorItem("."),
                queries.LiteralItem("abc"),
                queries.SubscriptOpenerItem("["),
                queries.LiteralItem("def[ghi"),
                queries.SubscriptCloserItem("]"),
                queries.LiteralItem("jkl]"),
                queries.QueryEndItem(),
            ],
        )

    def test_init_errors(self):
        """Test initialization errors"""
        for separator, subscript_pair, expected_exception_regex in (
            ("=", "[]", r"\ASeparator codepoint \d+ \('='\) not allowed"),
            (
                ".",
                "'",
                r"""\ASubscript indicator codepoint \d+ \("'"\) not allowed""",
            ),
            ("/", r"\/", r"\ASeparator codepoint \d+ \('/'\) not allowed"),
        ):
            with self.subTest(
                separator=separator,
                subscript_pair=subscript_pair,
            ):
                self.assertRaisesRegex(
                    ValueError,
                    expected_exception_regex,
                    queries.Itemizer,
                    separator_codepoint=ord(separator),
                    subscript_indicators_pair=commons.CharactersPair(
                        subscript_pair
                    ),
                )
            #
        #

    def test_invalid_components(self):
        """Test invalid query / path components"""
        for invalid_query, expected_exception_regex in (
            (".abc.def.{}", r"\A\{\} is not a valid path component"),
            (
                ".abc.{d,e,f}",
                r"\A\{'d': None, 'e': None, 'f': None\} is not"
                " a valid path component",
            ),
        ):
            with self.subTest(query=invalid_query):
                self.assertRaisesRegex(
                    TypeError,
                    expected_exception_regex,
                    queries.Parser().parse_query,
                    invalid_query,
                )
            #
        #


class ParsedQuery(TestCase):

    """ParsedQuery class"""

    def test_bool(self):
        """__bool__() special method"""
        for path, replacement, expected_result in (
            (TraversalPath(), Tree(1), True),
            (TraversalPath("abc"), Tree(1), True),
            (TraversalPath(), Tree(None), True),
            (TraversalPath("abc"), None, True),
            (TraversalPath(), None, False),
        ):
            with self.subTest(path=path, replacement=replacement):
                if expected_result:
                    self.assertTrue(queries.ParsedQuery(path, replacement))
                else:
                    self.assertFalse(queries.ParsedQuery(path, replacement))
                #
            #
        #

    def test_apply_to(self):
        """apply_to()  method"""
        for path, replacement, source_tree, expected_result in (
            (
                TraversalPath("abc"),
                Tree("def"),
                Tree(MapNode(abc=1, xyz=77)),
                Tree(MapNode(abc="def", xyz=77)),
            ),
            (
                TraversalPath(),
                Tree(MapNode(abc="def")),
                Tree(MapNode(abc=1, xyz=77)),
                Tree(MapNode(abc="def")),
            ),
            (
                TraversalPath("abc"),
                Tree(None),
                Tree(MapNode(abc=1, xyz=77)),
                Tree(MapNode(abc=None, xyz=77)),
            ),
            (
                TraversalPath("abc"),
                None,
                Tree(MapNode(abc=1, xyz=77)),
                Tree(1),
            ),
            (
                TraversalPath(),
                None,
                Tree(MapNode(abc=1, xyz=77)),
                Tree(MapNode(abc=1, xyz=77)),
            ),
        ):
            with self.subTest(path=path, replacement=replacement):
                parsed_query = queries.ParsedQuery(path, replacement)
                self.assertEqual(
                    parsed_query.apply_to(source_tree),
                    expected_result,
                )
            #
        #


class Parser(TestCase):

    """Test the Parser class"""

    def test_error_empty_query(self):
        """Test IllegalState error being raised"""
        parser = queries.Parser()
        self.assertRaisesRegex(
            queries.IllegalStateError,
            "The query must always start with a separator item!",
            parser.parse_query,
            "",
        )

    def test_working_queries(self):
        """Test working queries"""
        parser = queries.Parser()
        for query, expected_path, expected_replacement, warning in (
            (
                ".",
                TraversalPath(),
                None,
                None,
            ),
            (
                " .",
                TraversalPath(),
                None,
                None,
            ),
            (
                ".abc",
                TraversalPath("abc"),
                None,
                None,
            ),
            (
                ".abc[ d e f g h ]",
                TraversalPath("abc", "defgh"),
                None,
                None,
            ),
            (
                ".abc[def[ghi]jkl]",
                TraversalPath("abc", "def[ghi", "jkl]"),
                None,
                "Possible error: Found subscript",
            ),
            (
                ".abc['yes'].xyz = [1, 2, 4]",
                TraversalPath("abc", "yes", "xyz"),
                Tree(ListNode([1, 2, 4])),
                None,
            ),
            (
                ".abc['yes'].xyz = ",
                TraversalPath("abc", "yes", "xyz"),
                Tree(None),
                "Assuming empty replacement value",
            ),
        ):
            with self.subTest(query=query):
                if warning:
                    with self.assertLogs(None, level="WARNING") as log_cm:
                        self.assertEqual(
                            parser.parse_query(query),
                            queries.ParsedQuery(
                                expected_path, expected_replacement
                            ),
                        )
                    self.assertIn(warning, log_cm.output[0])
                else:
                    self.assertEqual(
                        parser.parse_query(query),
                        queries.ParsedQuery(
                            expected_path, expected_replacement
                        ),
                    )
                #
            #
        #

    def test_separator_variations(self):
        """Test working queries"""
        parser = queries.Parser()
        for query, expected_path, expected_replacement, separator in (
            (
                "/",
                TraversalPath(),
                None,
                "/",
            ),
            (
                " ~",
                TraversalPath(),
                None,
                "~",
            ),
            (
                "§abc",
                TraversalPath("abc"),
                None,
                "§",
            ),
            (
                "|abc | d e f g h ",
                TraversalPath("abc", "defgh"),
                None,
                "|",
            ),
            (
                ":abc['yes']:xyz = [1, 2, 4]",
                TraversalPath("abc", "yes", "xyz"),
                Tree(ListNode([1, 2, 4])),
                ":",
            ),
        ):
            with self.subTest(query=query, separator=separator):
                self.assertEqual(
                    parser.parse_query(
                        query, separator_codepoint=ord(separator)
                    ),
                    queries.ParsedQuery(expected_path, expected_replacement),
                )
            #
        #

    def test_illegal_states(self):
        """Test illegal states"""
        for mock_items, expected_error_regex in (
            (
                # before start
                [queries.LiteralItem("x")],
                "(?s)Parser phase: 'before start'.+"
                r"Expected QueryStartItem\(\)",
            ),
            (
                # started
                [
                    queries.QueryStartItem(),
                    queries.LiteralItem("x"),
                ],
                "(?s)Parser phase: 'started'.+"
                "The query must always start with a separator item",
            ),
            (
                # address
                [
                    queries.QueryStartItem(),
                    queries.SeparatorItem("."),
                    queries.LiteralItem("x"),
                    queries.QueryStartItem(),
                ],
                "(?s)Parser phase: 'address'.+"
                r"QueryStartItem\(\) not allowed here",
            ),
            (
                # replacement
                [
                    queries.QueryStartItem(),
                    queries.SeparatorItem("."),
                    queries.LiteralItem("x"),
                    queries.AssignmentItem("="),
                    queries.QueryStartItem(),
                ],
                "(?s)Parser phase: 'replacement'.+"
                r"QueryStartItem\(\) not allowed here",
            ),
            (
                # ended
                [
                    queries.QueryStartItem(),
                    queries.SeparatorItem("."),
                    queries.LiteralItem("x"),
                    queries.AssignmentItem("="),
                    queries.QueryEndItem(),
                    queries.WhitespaceItem(" "),
                ],
                "(?s)Parser phase: 'ended'.+No items allowed anymore",
            ),
        ):
            with self.subTest(expected_error_regex=expected_error_regex):
                with patch.object(
                    queries.Itemizer, "itemize", return_value=mock_items
                ):
                    parser = queries.Parser()
                    self.assertRaisesRegex(
                        queries.IllegalStateError,
                        expected_error_regex,
                        parser.parse_query,
                        "…",
                    )
                #
            #
        #

    @patch.object(queries.Parser, "phase", new="not implemented")
    def test_phase_not_implemented(self):
        """Test illegal states"""
        parser = queries.Parser()
        self.assertRaisesRegex(
            queries.IllegalStateError,
            "(?s)Parser phase: 'not implemented'.+"
            "Item handler for this phase not implemented",
            parser.parse_query,
            "…",
        )


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-

"""

dryjq.commandline

Command line functionality

Copyright (C) 2022 Rainer Schwarzbach

This file is part of dryjq.

dryjq is free software: you can redistribute it and/or modify
it under the terms of the MIT License.

dryjq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the LICENSE file for more details.

"""


import argparse
import io
import logging

import sys

from typing import Dict, List, Optional, TextIO, Type

import yaml

from serializable_trees import Tree

import dryjq

from dryjq import commons
from dryjq import streams

from dryjq.queries import ParsedQuery


#
# Constants
#


RETURNCODE_OK = 0
RETURNCODE_ERROR = 1


#
# classes
#


class Program:

    """Command line program"""

    name: str = "dryjq"
    description: str = "Drastically Reduced YAML / JSON Query"
    query_option: bool = True
    merge_program: bool = False
    modify_in_place_option: bool = True
    default_output_format: str = commons.FORMAT_INPUT

    allowed_formats: Dict[str, str] = {
        commons.FORMAT_JSON: "set output format to JSON",
        commons.FORMAT_YAML: "set output fromat to YAML",
        commons.FORMAT_INPUT: "keep input format",
        commons.FORMAT_TOGGLE: "change JSON to YAML or vice versa",
    }

    def __init__(self, args: Optional[List[str]] = None) -> None:
        """Parse command line arguments and initialize the logger

        :param args: a list of command line arguments
        """
        self.__arguments = self._parse_args(args)
        if self.__arguments.loglevel < logging.INFO:
            message_format = (
                "%(levelname)-8s | (%(funcName)s:%(lineno)s) %(message)s"
            )
        else:
            message_format = "%(levelname)-8s | %(message)s"
        #
        if self.query_option and self.__arguments.query is None:
            self.__arguments.query = self.__arguments.separator
        #
        logging.basicConfig(
            format=message_format,
            level=self.__arguments.loglevel,
        )

    @property
    def arguments(self) -> argparse.Namespace:
        """Property: command line arguments

        :returns: the parsed command line arguments
        """
        return self.__arguments

    @classmethod
    def complete_format(cls, input_format: str) -> str:
        """Lookup the matching format

        :param input_format: the (case insensitive) start of
            any allowed input format
        :returns: the matching format
        :raises: ValueError if the no matching format was found
        """
        lower_input_format = input_format.lower()
        for candidate in cls.allowed_formats:
            if candidate.lower().startswith(lower_input_format):
                return candidate
            #
        #
        raise ValueError(f"Unsupported format {input_format!r}!")

    def _parse_args(self, args: Optional[List[str]]) -> argparse.Namespace:
        """Parse command line arguments using argparse
        and return the arguments namespace.

        :param args: a list of command line arguments,
            or None to parse sys.argv
        :returns: the parsed command line arguments as returned
            by argparse.ArgumentParser().parse_args()
        """
        main_parser = argparse.ArgumentParser(
            prog=self.name,
            description=self.description,
        )
        main_parser.set_defaults(
            loglevel=logging.WARNING,
            output_format=self.default_output_format,
            indent=commons.DEFAULT_INDENT,
            inplace=False,
            separator=commons.DEFAULT_SEPARATOR,
            subscript_indicators=commons.DEFAULT_SUBSCRIPT_INDICATORS,
        )
        main_parser.add_argument(
            "--version",
            action="version",
            version=dryjq.__version__,
            help="print version and exit",
        )
        if self.modify_in_place_option:
            main_parser.add_argument(
                "-i",
                "--inplace",
                action="store_true",
                help="modify the input file in place instead of writing"
                " the result to standard output",
            )
        #
        logging_group = main_parser.add_argument_group(
            "Logging options", "control log level (default is WARNING)"
        )
        verbosity = logging_group.add_mutually_exclusive_group()
        verbosity.add_argument(
            "-d",
            "--debug",
            action="store_const",
            const=logging.DEBUG,
            dest="loglevel",
            help="output all messages (log level DEBUG)",
        )
        verbosity.add_argument(
            "-v",
            "--verbose",
            action="store_const",
            const=logging.INFO,
            dest="loglevel",
            help="be more verbose (log level INFO)",
        )
        verbosity.add_argument(
            "-q",
            "--quiet",
            action="store_const",
            const=logging.ERROR,
            dest="loglevel",
            help="be more quiet (log level ERROR)",
        )
        output_group = main_parser.add_argument_group(
            "Output options", "control output formatting"
        )
        format_choices = ", ".join(
            f"{key!r}: {value}"
            for (key, value) in self.allowed_formats.items()
        )
        output_group.add_argument(
            "-o",
            "--output-format",
            type=self.complete_format,
            help=f"output format (choice of {format_choices};"
            " default: %(default)r)",
        )
        output_group.add_argument(
            "--indent",
            choices=(2, 4, 8),
            type=int,
            help="indentation depth of blocks, in spaces"
            " (default: %(default)s)",
        )
        output_group.add_argument(
            "--sort-keys",
            action="store_true",
            help="sort mapping keys"
            " (by default, mapping keys are left in input order)",
        )
        if self.query_option:
            query_syntax_group = main_parser.add_argument_group(
                "Query syntax options", "control the query syntax"
            )
            query_syntax_group.add_argument(
                "--separator",
                help="the separator character (default: %(default)r)",
            )
            query_syntax_group.add_argument(
                "--subscript-indicators",
                help="the subscript indicator character(s)"
                " (default: %(default)r)",
            )
            main_parser.add_argument(
                "query",
                nargs="?",
                help="the query (simplest form of yq/jq syntax,"
                " default is the separator character alone).",
            )
        #
        if self.merge_program:
            main_parser.add_argument(
                "input_file",
                help="the input file name",
            )
            main_parser.add_argument(
                "merge_files",
                nargs="*",
                help="the names of the files to be merged"
                " (if no file name is provided here,"
                " one file will be read from standard input)",
            )
        else:
            main_parser.add_argument(
                "input_file",
                nargs="?",
                help="the input file name"
                " (by default, data will be read from standard input)",
            )
        #
        return main_parser.parse_args(args)

    def _do_merges(self, data_handler: streams.StreamReader) -> None:
        """Merge each file provided as positional parameters
        after the input file into the data structure made from
        the input file (data_handler will be modified in place)

        :param data_handler: the streams.StreamReader
            (or subclass) instance fro the input file
        """
        if not self.arguments.merge_files:
            logging.info("- ↓↓↓ Reading merge data from standard input")
            updating_data_structure = Tree.from_yaml(sys.stdin.read())
            logging.info("----- Finished reading merge data")
            data_handler.execute_merge(updating_data_structure)
            return
        #
        for merge_file_name in self.arguments.merge_files:
            logging.info(
                "----- Reading merge data from file: %s", merge_file_name
            )
            updating_data_structure = Tree.from_file(merge_file_name)
            data_handler.execute_merge(updating_data_structure)
        #

    def _write_output_data(
        self,
        handler_class: Type[streams.StreamReader],
        stream: TextIO,
        parsed_query: Optional[ParsedQuery] = None,
    ) -> int:
        """Execute the query using the data path,
        write output and return the returncode.
        Catch all expected exceptions and log them using ERROR loglevel.

        :param handler_class: streams.StreamReader (or a subclass of it)
        :param stream: the (text) IO stream containing the
            serialized data structure
        :param data_path: the TraversalPath constructed
            from the original query
        :param replacement: the Tree constructed
            from the original query equation right side
        :returns: the returncode for the script
        """
        try:
            stream_handler = handler_class(stream)
        except (yaml.YAMLError, yaml.composer.ComposerError) as error:
            for line in str(error).splitlines():
                logging.error(line)
            #
            return RETURNCODE_ERROR
        #
        try:
            if self.merge_program:
                self._do_merges(stream_handler)
            elif not parsed_query:
                logging.debug("Skipping empty query")
            else:
                stream_handler.execute_single_query(parsed_query)
            #
            stream_handler.set_serialization_format(
                output_format=self.arguments.output_format,
                indent=self.arguments.indent,
                sort_keys=self.arguments.sort_keys,
            )
            stream_handler.write_output()
        except (TypeError, ValueError, yaml.YAMLError) as error:
            for line in str(error).splitlines():
                logging.error(line)
            #
            return RETURNCODE_ERROR
        #
        return RETURNCODE_OK

    def execute(
        self,
        parsed_query: Optional[ParsedQuery] = None,
    ) -> int:
        """Execute the program

        :param data_path: the TraversalPath constructed
            from the original query
        :param replacement: the Tree constructed
            from the original query equation right side
        :returns: the returncode for the script
        """
        display_mode = "extract mode"
        replace_mode = (
            parsed_query is not None and parsed_query.replacement is not None
        )
        if replace_mode:
            display_mode = "replace mode"
        elif not parsed_query:
            display_mode = f"{display_mode} (passthrough)"
        #
        stream_handler_class = streams.StreamReader
        logging.info("Operating in %s.", display_mode)
        if self.arguments.input_file is None:
            if self.arguments.inplace:
                logging.warning("Cannot modify <stdin> in place")
            #
            logging.info("- ↓↓↓ Reading data from standard input")
            return self._write_output_data(
                stream_handler_class,
                sys.stdin,
                parsed_query=parsed_query,
            )
        #
        file_open_mode = "r"
        if replace_mode and self.arguments.inplace:
            file_open_mode = "r+"
            stream_handler_class = streams.FileWriter
        #
        with open(
            self.arguments.input_file,
            mode=file_open_mode,
            encoding="utf-8",
        ) as input_file:
            logging.info(
                "----- Reading data from file: %s", self.arguments.input_file
            )
            # Help mypy to be strict about the file type (TextIO)
            if isinstance(input_file, io.TextIOBase):
                returncode = self._write_output_data(
                    stream_handler_class,
                    input_file,
                    parsed_query=parsed_query,
                )
            else:
                logging.error("Not a text file!")  # NOT TESTABLE
                return RETURNCODE_ERROR  # NOT TESTABLE
            #
        #
        return returncode


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-

"""

test_streams

Unit test the streams module

Copyright (C) 2022 Rainer Schwarzbach

This file is part of dryjq.

dryjq is free software: you can redistribute it and/or modify
it under the terms of the MIT License.

dryjq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the LICENSE file for more details.

"""


import io
import os
import secrets
import sys
import tempfile
import time

from unittest import TestCase
from unittest.mock import patch

from serializable_trees import TraversalPath, Tree

from dryjq import queries
from dryjq import streams

from . import data


class StreamReader(TestCase):

    """StreamReader tests"""

    def test_yaml_properties(self):
        """Test properties (YAML input)"""
        stream_io = io.StringIO(data.INPUT_YAML_ANIMALS)
        stream_reader = streams.StreamReader(stream_io)
        stream_reader.execute_single_query(
            queries.ParsedQuery(
                TraversalPath("felines", "cats", "big", 1), None
            )
        )
        for tested_property, expected_value in (
            ("input_contents", data.INPUT_YAML_ANIMALS),
            ("input_format", "YAML"),
            ("output_contents", data.EXPECT_CATS_BIG_1),
            ("output_format", "YAML"),
            ("stream_io", stream_io),
        ):
            with self.subTest(
                tested_property=tested_property, expected_value=expected_value
            ):
                self.assertEqual(
                    getattr(stream_reader, tested_property),
                    expected_value,
                )
            #
        #

    def test_json_properties(self):
        """Test properties (JSON input)"""
        stream_io = io.StringIO(data.INPUT_JSON_ANIMALS)
        stream_reader = streams.StreamReader(stream_io)
        stream_reader.execute_single_query(
            queries.ParsedQuery(TraversalPath("canines"), None)
        )
        for tested_property, expected_value in (
            ("input_contents", data.INPUT_JSON_ANIMALS),
            ("input_format", "JSON"),
            ("output_contents", data.EXPECT_JSON_CANINES),
            ("output_format", "JSON"),
            ("stream_io", stream_io),
        ):
            with self.subTest(
                tested_property=tested_property, expected_value=expected_value
            ):
                self.assertEqual(
                    getattr(stream_reader, tested_property),
                    expected_value,
                )
            #
        #

    def test_json2yaml_properties(self):
        """Test properties (JSON input -> YAML output)"""
        stream_io = io.StringIO(data.INPUT_JSON_ANIMALS)
        stream_reader = streams.StreamReader(stream_io)
        stream_reader.execute_single_query(
            queries.ParsedQuery(TraversalPath("canines"), None)
        )
        stream_reader.set_serialization_format(output_format="YAML")
        for tested_property, expected_value in (
            ("input_contents", data.INPUT_JSON_ANIMALS),
            ("input_format", "JSON"),
            ("output_contents", data.EXPECT_YAML_CANINES),
            ("output_format", "YAML"),
            ("stream_io", stream_io),
        ):
            with self.subTest(
                tested_property=tested_property, expected_value=expected_value
            ):
                self.assertEqual(
                    getattr(stream_reader, tested_property),
                    expected_value,
                )
            #
        #

    def test_yaml2json_properties(self):
        """Test properties (YAML input -> JSON output)"""
        stream_io = io.StringIO(data.INPUT_YAML_ANIMALS)
        stream_reader = streams.StreamReader(stream_io)
        stream_reader.execute_single_query(
            queries.ParsedQuery(TraversalPath("canines"), None)
        )
        stream_reader.set_serialization_format(output_format="JSON")
        for tested_property, expected_value in (
            ("input_contents", data.INPUT_YAML_ANIMALS),
            ("input_format", "YAML"),
            ("output_contents", data.EXPECT_JSON_CANINES),
            ("output_format", "JSON"),
            ("stream_io", stream_io),
        ):
            with self.subTest(
                tested_property=tested_property, expected_value=expected_value
            ):
                self.assertEqual(
                    getattr(stream_reader, tested_property),
                    expected_value,
                )
            #
        #

    def test_valid_formats(self):
        """Test valid formats (YAML input)"""
        stream_io = io.StringIO(data.INPUT_YAML_ANIMALS)
        stream_reader = streams.StreamReader(stream_io)
        for output_format, expected_value in (
            ("input", data.INPUT_YAML_ANIMALS),
            ("toggle", data.EXPECT_CONVERTED_JSON_ANIMALS.rstrip()),
            ("YAML", data.INPUT_YAML_ANIMALS),
            ("JSON", data.EXPECT_CONVERTED_JSON_ANIMALS.rstrip()),
        ):
            with self.subTest(output_format=output_format):
                stream_reader.set_serialization_format(
                    output_format=output_format
                )
                self.assertEqual(
                    stream_reader.output_contents,
                    expected_value,
                )
            #
        #

    def test_unsupported_output_format(self):
        """Test unsupported output format"""
        stream_io = io.StringIO(data.INPUT_YAML_ANIMALS)
        stream_reader = streams.StreamReader(stream_io)
        self.assertRaisesRegex(
            ValueError,
            "Output format 'XML' not supported!",
            stream_reader.set_serialization_format,
            output_format="XML",
        )

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_writing_output(self, mock_stdout):
        """Test writing output to standard output"""
        assert mock_stdout is sys.stdout
        stream_io = io.StringIO(data.INPUT_JSON_ANIMALS)
        stream_reader = streams.StreamReader(stream_io)
        stream_reader.execute_single_query(
            queries.ParsedQuery(TraversalPath("canines"), None)
        )
        stream_reader.set_serialization_format(output_format="YAML")
        stream_reader.write_output()
        self.assertEqual(mock_stdout.getvalue(), data.EXPECT_YAML_CANINES)


class FileWriter(TestCase):

    """FileWriter tests"""

    def test_file_output(self):
        """Test modification of an existing file"""
        with tempfile.TemporaryDirectory() as temporary_directory:
            yaml_file_name = os.path.join(
                temporary_directory, f"{secrets.token_urlsafe(6)}.yaml"
            )
            with open(
                yaml_file_name, mode="w", encoding="utf-8"
            ) as source_file:
                source_file.write(data.INPUT_YAML_ANIMALS)
            #
            time.sleep(0.1)
            creation_time = os.stat(yaml_file_name).st_mtime
            with open(yaml_file_name, mode="r+", encoding="utf-8") as io_file:
                file_writer = streams.FileWriter(io_file)
                file_writer.execute_single_query(
                    queries.ParsedQuery(
                        TraversalPath("felines", "cats", "big", 2),
                        replacement=Tree("panther"),
                    )
                )
                file_writer.write_output()
            #
            # File changed after creation
            self.assertGreater(os.stat(yaml_file_name).st_mtime, creation_time)
            with open(
                yaml_file_name, mode="r", encoding="utf-8"
            ) as changed_file:
                self.assertEqual(
                    changed_file.read(), data.EXPECT_YAML_JAGUAR2PANTHER
                )
            #
        #

    def test_unchanged_output(self):
        """Test modification of an existing file without changing output"""
        with tempfile.TemporaryDirectory() as temporary_directory:
            yaml_file_name = os.path.join(
                temporary_directory, f"{secrets.token_urlsafe(6)}.yaml"
            )
            with open(
                yaml_file_name, mode="w", encoding="utf-8"
            ) as source_file:
                source_file.write(data.INPUT_YAML_ANIMALS)
            #
            time.sleep(0.1)
            creation_time = os.stat(yaml_file_name).st_mtime
            with open(yaml_file_name, mode="r+", encoding="utf-8") as io_file:
                file_writer = streams.FileWriter(io_file)
                with self.assertLogs(level="ERROR") as log_cm:
                    file_writer.write_output()
                #
                self.assertIn(
                    "Not writing file: contents did not change.",
                    log_cm.records[0].message,
                )
            #
            # File NOT changed after creation
            self.assertEqual(os.stat(yaml_file_name).st_mtime, creation_time)
            with open(
                yaml_file_name, mode="r", encoding="utf-8"
            ) as changed_file:
                self.assertEqual(changed_file.read(), data.INPUT_YAML_ANIMALS)
            #
        #

    def test_changed_format(self):
        """Test modification of an existing file with changed format"""
        with tempfile.TemporaryDirectory() as temporary_directory:
            yaml_file_name = os.path.join(
                temporary_directory, f"{secrets.token_urlsafe(6)}.yaml"
            )
            with open(
                yaml_file_name, mode="w", encoding="utf-8"
            ) as source_file:
                source_file.write(data.INPUT_YAML_ANIMALS)
            #
            time.sleep(0.1)
            creation_time = os.stat(yaml_file_name).st_mtime
            with open(yaml_file_name, mode="r+", encoding="utf-8") as io_file:
                file_writer = streams.FileWriter(io_file)
                file_writer.execute_single_query(
                    queries.ParsedQuery(
                        TraversalPath("felines", "cats", "big", 2),
                        replacement=Tree("panther"),
                    )
                )
                file_writer.set_serialization_format(output_format="JSON")
                with self.assertLogs(level="ERROR") as log_cm:
                    file_writer.write_output()
                #
                self.assertIn(
                    "Not writing file: data format changed.",
                    log_cm.records[0].message,
                )
            #
            # File NOT changed after creation
            self.assertEqual(os.stat(yaml_file_name).st_mtime, creation_time)
            with open(
                yaml_file_name, mode="r", encoding="utf-8"
            ) as changed_file:
                self.assertEqual(changed_file.read(), data.INPUT_YAML_ANIMALS)
            #
        #


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-

"""

test.test_main_stdin

Unit test the command line interface
with input from stdin

Copyright (C) 2022 Rainer Schwarzbach

This file is part of dryjq.

dryjq is free software: you can redistribute it and/or modify
it under the terms of the MIT License.

dryjq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the LICENSE file for more details.

"""


import io
import sys

from unittest import TestCase
from unittest.mock import patch

import dryjq

from dryjq import __main__

from . import data
from .commons import GenericCallResult


class CommandLineResult(GenericCallResult):

    """Command line call result"""

    @classmethod
    def do_call(cls, *args, **kwargs):
        """Do the real function call"""
        del kwargs
        return __main__.main(list(args))


class TestSimple(TestCase):

    """Simple test module using patched stdin and stdout"""

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_get_single_value(self, mock_stdout):
        """Get a single value"""
        result = CommandLineResult.from_call(
            ".felines.cats.big[1]",
            stdin_data=data.INPUT_YAML_ANIMALS,
            stdout=mock_stdout,
        )
        self.assertEqual(result.stdout, data.EXPECT_CATS_BIG_1)
        self.assertEqual(result.returncode, data.RETURNCODE_OK)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_get_subtree(self, mock_stdout):
        """Get a subtree"""
        result = CommandLineResult.from_call(
            ".canines",
            stdin_data=data.INPUT_YAML_ANIMALS,
            stdout=mock_stdout,
        )
        self.assertEqual(result.stdout, data.EXPECT_YAML_CANINES)
        self.assertEqual(result.returncode, data.RETURNCODE_OK)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_missing_list_value(self, mock_stdout):
        """Try to get a missing list value"""
        missing_index = 5
        with self.assertLogs(None, level="ERROR") as log_cm:
            result = CommandLineResult.from_call(
                f".felines.cats.big[{missing_index}]",
                stdin_data=data.INPUT_YAML_ANIMALS,
                stdout=mock_stdout,
            )
        #
        self.assertIn(
            f"TraversalPath('felines', 'cats', 'big', {missing_index!r})"
            " not found!",
            log_cm.output[0],
        )
        self.assertEqual(result.stdout, "")
        self.assertEqual(result.returncode, data.RETURNCODE_ERROR)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_invalid_query(self, mock_stdout):
        """Try to parse a malformed query"""
        with self.assertLogs(None, level="ERROR") as log_cm:
            result = CommandLineResult.from_call(
                "invalid query",
                stdin_data=data.INPUT_YAML_ANIMALS,
                stdout=mock_stdout,
            )
        #
        self.assertIn(
            "The query must always start with a separator item",
            log_cm.output[0],
        )
        self.assertEqual(result.stdout, "")
        self.assertEqual(result.returncode, data.RETURNCODE_ERROR)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_missing_mapping_value(self, mock_stdout):
        """Try to get a missing mapping value"""
        missing_key = "rodents"
        with self.assertLogs(None, level="ERROR") as log_cm:
            result = CommandLineResult.from_call(
                f".{missing_key}",
                stdin_data=data.INPUT_YAML_ANIMALS,
                stdout=mock_stdout,
            )
        #
        self.assertIn(
            f"TraversalPath({missing_key!r}) not found!",
            log_cm.output[0],
        )
        self.assertEqual(result.stdout, "")
        self.assertEqual(result.returncode, data.RETURNCODE_ERROR)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_indexing_scalar_value(self, mock_stdout):
        """Try to get an indexed value from a scalar"""
        wrong_index = 1
        query = ".felines.cats.big[0]"
        # query_parser = queries.Parser()
        # tree = Tree.from_yaml(data.INPUT_YAML_ANIMALS)
        # path = query_parser.parse_query(query)[0]
        # scalar_value = tree.get_native_item(path)
        with self.assertLogs(None, level="ERROR") as log_cm:
            result = CommandLineResult.from_call(
                f"{query}[{wrong_index}]",
                stdin_data=data.INPUT_YAML_ANIMALS,
                stdout=mock_stdout,
            )
        #
        self.assertIn(
            "Cannot traverse through a leaf",
            log_cm.output[0],
        )
        self.assertEqual(result.stdout, "")
        self.assertEqual(result.returncode, data.RETURNCODE_ERROR)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_key_indexing_list(self, mock_stdout):
        """Try to get a value indexed by a key from a list"""
        wrong_index = "lion"
        query = f".felines.cats.big[{wrong_index}]"
        with self.assertLogs(None, level="ERROR") as log_cm:
            result = CommandLineResult.from_call(
                f"{query}",
                stdin_data=data.INPUT_YAML_ANIMALS,
                stdout=mock_stdout,
            )
        #
        self.assertIn(
            "list indices must be integers or slices, not str",
            log_cm.output[0],
        )
        self.assertEqual(result.stdout, "")
        self.assertEqual(result.returncode, data.RETURNCODE_ERROR)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_invalid_yaml(self, mock_stdout):
        """Invalid input data"""
        with self.assertLogs(None, level="ERROR") as log_cm:
            result = CommandLineResult.from_call(
                ".",
                stdin_data="a: b: c: d:",
                stdout=mock_stdout,
            )
        #
        self.assertIn(
            # FIXME: PyYAML specific error message, might change
            "mapping values are not allowed here",
            log_cm.output[0],
        )
        self.assertEqual(result.stdout, "")
        self.assertEqual(result.returncode, data.RETURNCODE_ERROR)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_modify_stdin(self, mock_stdout):
        """Try to modify stdin in place"""
        assert mock_stdout is sys.stdout
        with self.assertLogs(None, level="WARNING") as log_cm:
            result = CommandLineResult.from_call(
                "--inplace",
                stdin_data=data.INPUT_YAML_ANIMALS,
                stdout=mock_stdout,
            )
        #
        self.assertIn("Cannot modify <stdin> in place", log_cm.output[0])
        self.assertEqual(result.returncode, data.RETURNCODE_OK)
        self.assertEqual(mock_stdout.getvalue(), data.INPUT_YAML_ANIMALS)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_replace_single_value(self, mock_stdout):
        """Replace a single value"""
        result = CommandLineResult.from_call(
            ". = 42",
            stdin_data="Answer to the Ultimate Question of Life,"
            " The Universe, and Everything",
            stdout=mock_stdout,
        )
        self.assertEqual(result.stdout, "42\n")
        self.assertEqual(result.returncode, data.RETURNCODE_OK)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_replace_item(self, mock_stdout):
        """Replace an item"""
        result = CommandLineResult.from_call(
            ".felines.cats.big[2] = panther",
            stdin_data=data.INPUT_YAML_ANIMALS,
            stdout=mock_stdout,
        )
        self.assertEqual(result.stdout, data.EXPECT_YAML_JAGUAR2PANTHER)
        self.assertEqual(result.returncode, data.RETURNCODE_OK)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_replace_subtree(self, mock_stdout):
        """Replace a subtree"""
        result = CommandLineResult.from_call(
            f".felines.cats = {data.REPLACEMENT_YAML_CATS_TAXONOMY}",
            stdin_data=data.INPUT_YAML_ANIMALS,
            stdout=mock_stdout,
        )
        self.assertEqual(result.stdout, data.EXPECT_YAML_CATS_TAXONOMY)
        self.assertEqual(result.returncode, data.RETURNCODE_OK)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_get_json_value(self, mock_stdout):
        """Get a single value from JSON input"""
        result = CommandLineResult.from_call(
            ".aurora.australis",
            stdin_data=data.INPUT_JSON_AURORA,
            stdout=mock_stdout,
        )
        self.assertEqual(result.stdout, data.EXPECT_JSON_AURORA_AUSTRALIS)
        self.assertEqual(result.returncode, data.RETURNCODE_OK)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_convert_json2yaml(self, mock_stdout):
        """Convert JSON to YAML"""
        result = CommandLineResult.from_call(
            "--output-format",
            "YAML",
            stdin_data=data.INPUT_JSON_AURORA,
            stdout=mock_stdout,
        )
        self.assertEqual(result.stdout, data.EXPECT_CONVERTED_YAML_AURORA)
        self.assertEqual(result.returncode, data.RETURNCODE_OK)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_convert_yaml2json(self, mock_stdout):
        """Convert YAML to json"""
        result = CommandLineResult.from_call(
            "--output-format",
            "JSON",
            stdin_data=data.INPUT_YAML_ANIMALS,
            stdout=mock_stdout,
        )
        self.assertEqual(result.stdout, data.EXPECT_CONVERTED_JSON_ANIMALS)
        self.assertEqual(result.returncode, data.RETURNCODE_OK)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_print_version(self, mock_stdout):
        """Print version"""
        stdout = mock_stdout
        with self.assertRaises(SystemExit) as sys_exit:
            __main__.main(["--version"])
        #
        self.assertEqual(sys_exit.exception.code, data.RETURNCODE_OK)
        self.assertEqual(stdout.getvalue().rstrip(), dryjq.__version__)


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-

"""

test.test_commandline

Unit test the commandline module

Copyright (C) 2022 Rainer Schwarzbach

This file is part of dryjq.

dryjq is free software: you can redistribute it and/or modify
it under the terms of the MIT License.

dryjq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the LICENSE file for more details.

"""


import io
import itertools
import os
import secrets
import tempfile
import time

from unittest import TestCase

from unittest.mock import patch

from serializable_trees import TraversalPath, Tree

from dryjq import commandline
from dryjq import queries

from . import data
from .commons import GenericCallResult


class ExecResult(GenericCallResult):

    """Program execution result"""

    @classmethod
    def do_call(cls, *args, **kwargs):
        """Do the real function call"""
        program = commandline.Program(list(args))
        return program.execute(parsed_query=kwargs.get("parsed_query"))


class Program(TestCase):

    """Test the Program class"""

    def test_complete_format(self):
        """Test completing a format, all combinations"""
        for target_format in ("input", "toggle", "JSON", "YAML"):
            combinations = [
                (char.lower(), char.upper()) for char in target_format
            ]
            matching_abbreviations = set()
            for index in range(len(target_format)):
                matching_abbreviations.update(
                    set(itertools.product(*combinations[: index + 1]))
                )
            #
            for current_input in matching_abbreviations:
                format_str = "".join(current_input)
                with self.subTest(format_str=format_str):
                    self.assertEqual(
                        commandline.Program.complete_format(format_str),
                        target_format,
                    )
                #
            #
        #

    def test_unsupported_format(self):
        """Test unsupported formats"""
        for target_format in ("XML", "xml", "ini", "toml"):
            with self.subTest(target_format=target_format):
                self.assertRaisesRegex(
                    ValueError,
                    f"Unsupported format {target_format!r}!",
                    commandline.Program.complete_format,
                    target_format,
                )
            #
        #
        #

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_exec_stdin_extract(self, mock_stdout):
        """Test execute function, reading stdin, extract mode"""
        with self.assertLogs(level="INFO") as log_cm:
            result = ExecResult.from_call(
                "--verbose",
                stdin_data=data.INPUT_YAML_ANIMALS,
                stdout=mock_stdout,
                parsed_query=queries.ParsedQuery(
                    TraversalPath("felines", "cats", "big", 1), None
                ),
            )
        self.assertEqual(
            log_cm.records[0].message, "Operating in extract mode."
        )
        self.assertEqual(result.stdout, data.EXPECT_CATS_BIG_1)
        self.assertEqual(result.returncode, data.RETURNCODE_OK)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_exec_stdin_replace_none_path(self, mock_stdout):
        """Test execute function, reading stdin, replace mode, None path"""
        with self.assertLogs(level="INFO") as log_cm:
            result = ExecResult.from_call(
                "--verbose",
                stdin_data=data.INPUT_YAML_ANIMALS,
                stdout=mock_stdout,
                parsed_query=queries.ParsedQuery(
                    TraversalPath(),
                    Tree("xyz"),
                ),
            )
        self.assertEqual(
            log_cm.records[0].message,
            "Operating in replace mode.",
        )
        self.assertEqual(result.stdout, "xyz\n")
        self.assertEqual(result.returncode, data.RETURNCODE_OK)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_exec_stdin_replace(self, mock_stdout):
        """Test execute function, reading stdin, replace mode"""
        with self.assertLogs(level="INFO") as log_cm:
            result = ExecResult.from_call(
                "--debug",
                stdin_data=data.INPUT_YAML_ANIMALS,
                stdout=mock_stdout,
                parsed_query=queries.ParsedQuery(
                    TraversalPath("felines", "cats", "big", 2),
                    Tree("panther"),
                ),
            )
        self.assertEqual(
            log_cm.records[0].message, "Operating in replace mode."
        )
        self.assertEqual(result.stdout, data.EXPECT_YAML_JAGUAR2PANTHER)
        self.assertEqual(result.returncode, data.RETURNCODE_OK)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_exec_stdin_passthru(self, mock_stdout):
        """Test execute function, reading stdin, extract mode (passthru)"""
        with self.assertLogs(level="INFO") as log_cm:
            result = ExecResult.from_call(
                "--verbose",
                stdin_data=data.INPUT_YAML_ANIMALS,
                stdout=mock_stdout,
            )
        #
        self.assertEqual(
            log_cm.records[0].message,
            "Operating in extract mode (passthrough).",
        )
        self.assertEqual(result.stdout, data.INPUT_YAML_ANIMALS)
        self.assertEqual(result.returncode, data.RETURNCODE_OK)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_exec_invalid_input(self, mock_stdout):
        """Test execute function, reading stdin, invalid input"""
        with self.assertLogs(level="ERROR") as log_cm:
            result = ExecResult.from_call(
                stdin_data="a: b: c",
                stdout=mock_stdout,
            )
        #
        self.assertIn(
            # FIXME: PyYAML specific error message, might change
            "mapping values are not allowed here",
            log_cm.records[0].message,
        )
        self.assertEqual(result.stdout, "")
        self.assertEqual(result.returncode, data.RETURNCODE_ERROR)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_exec_not_found(self, mock_stdout):
        """Test execute function, reading stdin, value not found"""
        wrong_index = "HUGE"
        with self.assertLogs(level="ERROR") as log_cm:
            result = ExecResult.from_call(
                stdin_data=data.INPUT_YAML_ANIMALS,
                stdout=mock_stdout,
                parsed_query=queries.ParsedQuery(
                    TraversalPath("felines", "cats", wrong_index),
                    None,
                ),
            )
        #
        self.assertEqual(
            log_cm.records[0].message,
            f"TraversalPath('felines', 'cats', {wrong_index!r}) not found!",
        )
        self.assertEqual(result.stdout.rstrip(), "")
        self.assertEqual(result.returncode, data.RETURNCODE_ERROR)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_exec_stdin_inplace(self, mock_stdout):
        """Test execute function, reading stdin, replace mode with --inplace"""
        with self.assertLogs(level="WARNING") as log_cm:
            result = ExecResult.from_call(
                "--inplace",
                stdin_data=data.INPUT_YAML_ANIMALS,
                stdout=mock_stdout,
                parsed_query=queries.ParsedQuery(
                    TraversalPath("felines", "cats", "big", 2),
                    Tree("panther"),
                ),
            )
        self.assertEqual(
            log_cm.records[0].message, "Cannot modify <stdin> in place"
        )
        self.assertEqual(result.stdout, data.EXPECT_YAML_JAGUAR2PANTHER)
        self.assertEqual(result.returncode, data.RETURNCODE_OK)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_exec_file_read(self, mock_stdout):
        """Test reading a file"""
        with tempfile.TemporaryDirectory() as temporary_directory:
            yaml_file_name = os.path.join(
                temporary_directory, f"{secrets.token_urlsafe(6)}.yaml"
            )
            with open(
                yaml_file_name, mode="w", encoding="utf-8"
            ) as source_file:
                source_file.write(data.INPUT_YAML_ANIMALS)
            #
            time.sleep(0.1)
            creation_time = os.stat(yaml_file_name).st_mtime
            result = ExecResult.from_call(
                "",
                yaml_file_name,
                stdout=mock_stdout,
                parsed_query=queries.ParsedQuery(
                    TraversalPath("felines", "cats", "big", 2),
                    Tree("panther"),
                ),
            )
            check_time = os.stat(yaml_file_name).st_mtime
        #
        self.assertEqual(result.stdout, data.EXPECT_YAML_JAGUAR2PANTHER)
        self.assertEqual(result.returncode, data.RETURNCODE_OK)
        self.assertEqual(check_time, creation_time)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_exec_file_write(self, mock_stdout):
        """Test modifying a file in place"""
        with tempfile.TemporaryDirectory() as temporary_directory:
            yaml_file_name = os.path.join(
                temporary_directory, f"{secrets.token_urlsafe(6)}.yaml"
            )
            with open(
                yaml_file_name, mode="w", encoding="utf-8"
            ) as source_file:
                source_file.write(data.INPUT_YAML_ANIMALS)
            #
            time.sleep(0.1)
            creation_time = os.stat(yaml_file_name).st_mtime
            result = ExecResult.from_call(
                "--inplace",
                "",
                yaml_file_name,
                stdout=mock_stdout,
                parsed_query=queries.ParsedQuery(
                    TraversalPath("felines", "cats", "big", 2),
                    Tree("panther"),
                ),
            )
            check_time = os.stat(yaml_file_name).st_mtime
            self.assertGreater(check_time, creation_time)
            with open(
                yaml_file_name, mode="r", encoding="utf-8"
            ) as changed_file:
                written_data = changed_file.read()
            #
            self.assertEqual(written_data, data.EXPECT_YAML_JAGUAR2PANTHER)
        #
        self.assertEqual(result.stdout, "")
        self.assertEqual(result.returncode, data.RETURNCODE_OK)


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-

"""

dryjq.commons

Common constants and helper classes

Copyright (C) 2022 Rainer Schwarzbach

This file is part of dryjq.

dryjq is free software: you can redistribute it and/or modify
it under the terms of the MIT License.

dryjq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the LICENSE file for more details.

"""

from typing import Tuple


DEFAULT_INDENT = 2
DEFAULT_SEPARATOR = "."
DEFAULT_SUBSCRIPT_INDICATORS = "[]"

FORMAT_INPUT = "input"
FORMAT_JSON = "JSON"
FORMAT_YAML = "YAML"
FORMAT_TOGGLE = "toggle"

SUPPORTED_FORMATS: Tuple[str, str] = (FORMAT_JSON, FORMAT_YAML)


#
# Classes
#


class CharactersPair:

    """Helper class keeping a pair of characters"""

    def __init__(self, source: str) -> None:
        """Store the codepoints of first and second character

        :param source: a string of 1 or 2 characters length
        :raises: ValueError if the string length exceeds 2,
            IndexError if an empty string was provided.
        """
        if len(source) > 2:
            raise ValueError("Please provide one or two characters only!")
        #
        self.__first_codepoint = ord(source[0])
        try:
            self.__last_codepoint = ord(source[1])
        except IndexError:
            self.__last_codepoint = self.__first_codepoint
        #
        self.__source = source

    @property
    def source(self) -> str:
        """Property: the source

        :returns: the source string
        """
        return self.__source

    @property
    def first(self) -> int:
        """Property: first codepoint

        :returns: the first codepoint
        """
        return self.__first_codepoint

    @property
    def last(self) -> int:
        """Property: last codepoint

        :returns: the last (=second) codepoint
        """
        return self.__last_codepoint

    @property
    def both(self) -> Tuple[int, int]:
        """Property: both codepoints

        :returns: a tuple containing both codepoints in order
        """
        return (self.first, self.last)

    # Aliases for the properies
    open = first
    close = last

    def __eq__(self, other) -> bool:
        """Comparison: self == other
        Focus on equal effect of the instances,
        not on exact equal initialization.

        :param other: another instance of the same class
        :returns: True if the 'both' property of both instances
            are equal, False otherwise
        """
        return self.both == other.both

    def __hash__(self) -> int:
        """Hash value computation: hash()

        :returns: the hash value of the source
        """
        return hash(self.source)


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-

"""

dryjq

Core module

Copyright (C) 2022-2023 Rainer Schwarzbach

This file is part of dryjq.

dryjq is free software: you can redistribute it and/or modify
it under the terms of the MIT License.

dryjq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the LICENSE file for more details.

"""

__version__ = "1.9.1"


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-

"""

tests.test_merge

Unit test the 'dryjq.merge' command line interface

Copyright (C) 2023 Rainer Schwarzbach

This file is part of dryjq.

dryjq is free software: you can redistribute it and/or modify
it under the terms of the MIT License.

dryjq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the LICENSE file for more details.

"""


import io
import os
import secrets
import tempfile

from unittest import TestCase
from unittest.mock import patch

from dryjq import merge

from . import data
from .commons import GenericCallResult


class ConvertResult(GenericCallResult):

    """Conversion result"""

    @classmethod
    def do_call(cls, *args, **kwargs):
        """Do the real function call"""
        del kwargs
        return merge.main(list(args))


class TestSimple(TestCase):

    """Simple test module using patched stdin and stdout"""

    def setUp(self):
        """Create a temporary directory"""
        # pylint: disable=consider-using-with
        self.tempdir = tempfile.TemporaryDirectory()
        # pylint: enable=consider-using-with
        self.testfile_name = os.path.join(
            self.tempdir.name, f"{secrets.token_urlsafe(6)}.yaml"
        )

    def tearDown(self):
        """Cleanup the temporary directory"""
        self.tempdir.cleanup()

    def __create_file(self, content=None):
        """Create a file in the temporary directory"""
        if content:
            with open(
                self.testfile_name,
                mode="w",
                encoding="utf-8",
            ) as testdata_file:
                testdata_file.write(content)
            #
        #

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_merge_via_stdin(self, mock_stdout):
        """Convert JSON to YAML"""
        self.__create_file(content=data.INPUT_YAML_ANIMALS)
        result = ConvertResult.from_call(
            self.testfile_name,
            stdin_data=data.INPUT_ANIMAL_MERGER_1,
            stdout=mock_stdout,
        )
        self.assertEqual(result.stdout, data.EXPECT_MERGED_ANIMALS_1)
        self.assertEqual(result.returncode, data.RETURNCODE_OK)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_merge_via_file(self, mock_stdout):
        """merge via file"""
        self.__create_file(content=data.INPUT_YAML_ANIMALS)
        merge_file_name = os.path.join(
            self.tempdir.name, f"to-merge-{secrets.token_urlsafe(6)}.yaml"
        )
        with open(
            merge_file_name,
            mode="w",
            encoding="utf-8",
        ) as testdata_file:
            testdata_file.write(data.INPUT_ANIMAL_MERGER_1)
        #
        result = ConvertResult.from_call(
            self.testfile_name,
            merge_file_name,
            stdin_data=data.INPUT_ANIMAL_MERGER_1,
            stdout=mock_stdout,
        )
        self.assertEqual(result.stdout, data.EXPECT_MERGED_ANIMALS_1)
        self.assertEqual(result.returncode, data.RETURNCODE_OK)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_print_version(self, mock_stdout):
        """Print version"""
        with self.assertRaises(SystemExit) as sys_exit:
            ConvertResult.from_call(
                "--version",
                stdout=mock_stdout,
            )
        #
        self.assertEqual(sys_exit.exception.code, data.RETURNCODE_OK)


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-

"""

dryjq.convert

dryjq.convert (or python3 -m dryjq.convert) command line script

Copyright (C) 2022 Rainer Schwarzbach

This file is part of dryjq.

dryjq is free software: you can redistribute it and/or modify
it under the terms of the MIT License.

dryjq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the LICENSE file for more details.

"""


import sys

from typing import List, Optional

from dryjq import commons
from dryjq import commandline


#
# Classes
#


class ConvertProgram(commandline.Program):

    """Conversion program
    commandline.Program subclass with less command line arguments
    and a different default output format compared to the parent class
    """

    name: str = "dryjq.convert"
    description: str = "Drastically Reduced YAML / JSON Query: convert files"
    query_option: bool = False
    merge_program: bool = False
    modify_in_place_option: bool = False
    default_output_format: str = commons.FORMAT_TOGGLE


#
# Functions
#


def main(args: Optional[List[str]] = None) -> int:
    """Execute the program using the provided command line arguments

    :param args: the list of command line arguments,
        or None to go with the default (sys.argv).
    :returns: the script returncode
    """
    program = ConvertProgram(args)
    return program.execute()


if __name__ == "__main__":
    sys.exit(main())  # NOT TESTABLE


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-

"""

test.test_convert

Unit test the 'dryjq.convert' command line interface

Copyright (C) 2022 Rainer Schwarzbach

This file is part of dryjq.

dryjq is free software: you can redistribute it and/or modify
it under the terms of the MIT License.

dryjq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the LICENSE file for more details.

"""


import io

from unittest import TestCase
from unittest.mock import patch

from dryjq import convert

from . import data
from .commons import GenericCallResult


class ConvertResult(GenericCallResult):

    """Conversion result"""

    @classmethod
    def do_call(cls, *args, **kwargs):
        """Do the real function call"""
        del kwargs
        return convert.main(list(args))


class TestSimple(TestCase):

    """Simple test module using patched stdin and stdout"""

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_convert_json2yaml(self, mock_stdout):
        """Convert JSON to YAML"""
        result = ConvertResult.from_call(
            stdin_data=data.INPUT_JSON_AURORA,
            stdout=mock_stdout,
        )
        self.assertEqual(result.stdout, data.EXPECT_CONVERTED_YAML_AURORA)
        self.assertEqual(result.returncode, data.RETURNCODE_OK)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_convert_yaml2json(self, mock_stdout):
        """Convert YAML to json"""
        result = ConvertResult.from_call(
            stdin_data=data.INPUT_YAML_ANIMALS,
            stdout=mock_stdout,
        )
        self.assertEqual(result.stdout, data.EXPECT_CONVERTED_JSON_ANIMALS)
        self.assertEqual(result.returncode, data.RETURNCODE_OK)

    @patch("sys.stdout", new_callable=io.StringIO)
    def test_print_version(self, mock_stdout):
        """Print version"""
        with self.assertRaises(SystemExit) as sys_exit:
            ConvertResult.from_call(
                "--version",
                stdout=mock_stdout,
            )
        #
        self.assertEqual(sys_exit.exception.code, data.RETURNCODE_OK)


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-

"""

dryjq.streams

Stream and file handlers

Copyright (C) 2022 Rainer Schwarzbach

This file is part of dryjq.

dryjq is free software: you can redistribute it and/or modify
it under the terms of the MIT License.

dryjq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the LICENSE file for more details.

"""


import json
import logging
import sys

from typing import TextIO

from serializable_trees import Tree

from dryjq import commons

from dryjq.queries import ParsedQuery


class StreamReader:

    """YAML or JSON stream reader and data handler.
    Can operate on all forms of (text) streams.
    """

    def __init__(self, stream_io: TextIO) -> None:
        """Read all data from the provided stream,
        detect if the format is JSON or YAML,
        and deserialize it into an internal data structure.

        :param stream_io: the stream to operate on.
        """
        if stream_io.seekable():
            stream_io.seek(0)
        #
        self.__input_contents = stream_io.read()
        logging.info("----- Finished reading data")
        try:
            self.__filtered_data = Tree.from_json(self.__input_contents)
        except json.JSONDecodeError:
            self.__filtered_data = Tree.from_yaml(self.__input_contents)
            self.__input_format = commons.FORMAT_YAML
        else:
            self.__input_format = commons.FORMAT_JSON
        #
        # self.__output_contents = self.__input_contents
        self.__output_format = self.__input_format
        self.__indent = commons.DEFAULT_INDENT
        self.__sort_keys = False
        self.__stream_io = stream_io

    @property
    def input_contents(self) -> str:
        """Property: the original contents read from the stream

        :returns: the original contents as a string.
        """
        return self.__input_contents

    @property
    def input_format(self) -> str:
        """Property: the detected input format

        :returns: the input format
        """
        return self.__input_format

    @property
    def output_contents(self) -> str:
        """Property: the internal data structure,
        serialized as JSON or YAML.

        :returns: the serialized data structure as a string
        """
        if self.__output_format == commons.FORMAT_JSON:
            return self.__filtered_data.to_json(
                indent=self.__indent,
                sort_keys=self.__sort_keys,
            )
        #
        output = self.__filtered_data.to_yaml(
            indent=self.__indent,
            sort_keys=self.__sort_keys,
        )
        if not isinstance(
            self.__filtered_data.root, (dict, list)
        ) and output.rstrip().endswith("\n..."):
            output = output.rstrip()[:-3]
        #
        return output

    @property
    def output_format(self) -> str:
        """Property: the output format

        :returns: the output format
        """
        return self.__output_format

    @property
    def stream_io(self) -> TextIO:
        """Property: the stream handle

        :returns: the TextIO instance of the stream
        """
        return self.__stream_io

    def set_serialization_format(
        self,
        output_format: str = commons.FORMAT_INPUT,
        indent: int = commons.DEFAULT_INDENT,
        sort_keys: bool = False,
    ) -> None:
        """Set the serialization format for the internal data structure.

        :param output_format: the output format
            (JSON, YAML, input or toggle).
        :param indent: the indent width.
        :param sort_keys: whether to sort mapping keys.
        """
        if output_format in commons.SUPPORTED_FORMATS:
            self.__output_format = output_format
        elif output_format == commons.FORMAT_INPUT:
            self.__output_format = self.input_format
        elif output_format == commons.FORMAT_TOGGLE:
            self.__output_format = commons.SUPPORTED_FORMATS[
                1 - commons.SUPPORTED_FORMATS.index(self.input_format)
            ]
        else:
            raise ValueError(f"Output format {output_format!r} not supported!")
        #
        self.__indent = indent
        self.__sort_keys = sort_keys

    def execute_merge(self, updating_data_structure: Tree) -> None:
        """merge updating_data_structure into the internal data structure.

        :param updating_data_structure: a data structure
            to be merged into self.__filtered_data
        """
        logging.debug("Filtered data before executing the merge:")
        logging.debug("%r", self.__filtered_data)
        self.__filtered_data = self.__filtered_data.joined_tree(
            updating_data_structure
        )
        logging.debug("Filtered data after executing the merge:")
        logging.debug("%r", self.__filtered_data)

    def execute_single_query(
        self,
        parsed_query: ParsedQuery,
    ) -> None:
        """Apply the provided Path to the internal data structure.

        :param access_path: an access.Path subclass instance
        """
        logging.debug("Filtered data before executing the query:")
        logging.debug("%r", self.__filtered_data)
        try:
            self.__filtered_data = parsed_query.apply_to(self.__filtered_data)
        except (IndexError, KeyError) as error:
            logging.debug(str(error))
            raise ValueError(f"{parsed_query.path} not found!") from error
        #
        logging.debug("Filtered data after executing the query:")
        logging.debug("%r", self.__filtered_data)

    def write_output(self) -> None:
        """Write output to stdout with a trailing Newline"""
        serialized_data = self.output_contents.rstrip()
        sys.stdout.write(f"{serialized_data}\n")


class FileWriter(StreamReader):

    """File writer.
    StreamReader subclass operating on text files opened
    for reading and writing (mode 'r+').
    """

    def write_output(self) -> None:
        """Write output to the file as is, but only if:
        - data has changed, and
        - the serialization format did not change.
        Log an error if one of these two hindsights applies.
        """
        not_writing = "Not writing file:"
        serialized_data = self.output_contents
        if serialized_data == self.input_contents:
            logging.error("%s contents did not change.", not_writing)
            return
        #
        if self.output_format != self.input_format:
            logging.error("%s data format changed.", not_writing)
            return
        #
        self.stream_io.seek(0)
        self.stream_io.truncate(0)
        self.stream_io.write(serialized_data)


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-

"""

dryjq.queries

Query parser and helper classes

Copyright (C) 2022 Rainer Schwarzbach

This file is part of dryjq.

dryjq is free software: you can redistribute it and/or modify
it under the terms of the MIT License.

dryjq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the LICENSE file for more details.

"""


import collections
import dataclasses
import logging

from typing import Iterator, List, Optional, Type

import yaml

from serializable_trees.basics import ScalarType, SCALAR_TYPES
from serializable_trees import TraversalPath, Tree

from dryjq import commons


#
# Item dataclasses
#


@dataclasses.dataclass(frozen=True)
class BaseItem:

    """Base class for all items"""


@dataclasses.dataclass(frozen=True)
class DelimiterItem(BaseItem):

    """Start or end item base class"""


@dataclasses.dataclass(frozen=True)
class QueryStartItem(DelimiterItem):

    """Query start item"""


@dataclasses.dataclass(frozen=True)
class QueryEndItem(DelimiterItem):

    """Query end item"""


@dataclasses.dataclass(frozen=True)
class ItemWithContent(BaseItem):

    """Base class for items with content"""

    content: str


@dataclasses.dataclass(frozen=True)
class AssignmentItem(ItemWithContent):

    """Assignment item"""


@dataclasses.dataclass(frozen=True)
class LiteralItem(ItemWithContent):

    """Literal item"""


@dataclasses.dataclass(frozen=True)
class SeparatorItem(ItemWithContent):

    """Separator item"""


@dataclasses.dataclass(frozen=True)
class SubscriptOpenerItem(ItemWithContent):

    """Subscript opener item"""


@dataclasses.dataclass(frozen=True)
class SubscriptCloserItem(ItemWithContent):

    """Subscript closer item"""


@dataclasses.dataclass(frozen=True)
class WhitespaceItem(ItemWithContent):

    """Whitespace item"""


#
# Exceptions
#


class IllegalStateError(Exception):

    """Exception raised if the parser is in an illegal state"""

    def __init__(
        self,
        message: str,
        offending_item: BaseItem,
        parser_phase: Optional[str] = None,
        position: int = 0,
    ) -> None:
        """Store the message, offending item, parser state and position

        :param message: the error details
        :param offending_item: the offending item
        :param parser_phase: the current parser phase
        :param position: the offending item’s position
        """
        super().__init__()
        self.message = message
        self.offending_item = offending_item
        self.parser_phase = parser_phase
        self.position = position

    def __str__(self) -> str:
        """String representation: str()

        :returns: the full error message
        """
        position_display = ""
        if self.position:
            position_display = (
                f" at item position #{self.position}, counting from 1"
            )
            if logging.getLogger().getEffectiveLevel() > logging.DEBUG:
                position_display = (
                    f"{position_display}; enable debug mode (-d)"
                    " for a list of determined items and their positions"
                )
            #
        #
        message = ""
        if self.message:
            message = f"\n{self.message}!"
        #
        return (
            f"Illegal state encountered{position_display}."
            f" Parser phase: {self.parser_phase!r}."
            f" Offending item: {self.offending_item!r}.{message}"
        )


class MalformedQueryError(Exception):

    """Exception raised by the itemizer on malformed queries"""

    error_type = "Malformed query"

    def __init__(self, character_position: int = 0) -> None:
        """Store the character position

        :param character_position: the character position
            which is significant for the error
        """
        super().__init__()
        self.character_position = character_position

    def __str__(self) -> str:
        """String representation: str()

        :returns: the full error message
        """
        return (
            f"{self.error_type} at character position"
            f" #{self.character_position} (counting from 1)"
        )


class UnclosedQuoteError(MalformedQueryError):

    """Exception raised when a quote was not closed"""

    error_type = "Unclosed quote started"


class UnclosedSubscriptError(MalformedQueryError):

    """Exception raised when a subscript was not closed"""

    error_type = "Unclosed subscript started"


#
# Classes
#


class Itemizer:

    """Itemize a query string for the parser"""

    quotes = (0x22, 0x27)
    equals_sign = 0x3D
    whitespace = (0x09, 0x0A, 0x0D, 0x20)

    def __init__(
        self,
        separator_codepoint: Optional[int] = None,
        subscript_indicators_pair: Optional[commons.CharactersPair] = None,
    ) -> None:
        """Initialize the itemizer internal state

        :param separator_codepoint: (optionally) the codepoint
            of the separator
        :param subscript_indicators_pair: (optionally) the subscript
            indicators pair
        :raises: ValueError if the separator codepoint or any of the
            subscript indicator codepoints are already taken,
            see the quotes, equals_sign and whitespace definitions
            in the class.
        """
        reserved_codepoints: set[int] = {
            self.equals_sign,
            *self.quotes,
            *self.whitespace,
        }
        if subscript_indicators_pair is None:
            subscript_indicators_pair = commons.CharactersPair(
                commons.DEFAULT_SUBSCRIPT_INDICATORS
            )
        #
        for subscript_codepoint in subscript_indicators_pair.both:
            if subscript_codepoint in reserved_codepoints:
                raise ValueError(
                    f"Subscript indicator codepoint {subscript_codepoint}"
                    f" ({chr(subscript_codepoint)!r}) not allowed,"
                    f" must not be any of {reserved_codepoints}!"
                )
            #
        #
        reserved_codepoints.update(subscript_indicators_pair.both)
        if separator_codepoint is None:
            separator_codepoint = ord(commons.DEFAULT_SEPARATOR)
        #
        if separator_codepoint in reserved_codepoints:
            raise ValueError(
                f"Separator codepoint {separator_codepoint}"
                f" ({chr(separator_codepoint)!r}) not allowed,"
                f" must not be any of {reserved_codepoints}!"
            )
        #
        self.__character_position = 0
        self.__current_literal: List[str] = []
        self.__items_queue: collections.deque[BaseItem] = collections.deque()
        self.__query_characters: collections.deque[str] = collections.deque()
        self.__separator_codepoint = separator_codepoint
        self.__subscript_indicators_pair = subscript_indicators_pair
        self.__subscript_start_position = 0

    @property
    def separator_codepoint(self) -> int:
        """Property: separator codepoint

        :returns: the separator codepoint
        """
        return self.__separator_codepoint

    @property
    def subscript_indicators_pair(self) -> commons.CharactersPair:
        """Property: subscript indicators

        :returns: the subscript indicators characters pair
        """
        return self.__subscript_indicators_pair

    def _add_current_literal(self) -> None:
        """Add a new LiteralItem from the current literal
        if any literal fragments have been collected yet.
        """
        if self.__current_literal:
            self.__items_queue.append(
                LiteralItem("".join(self.__current_literal))
            )
            self.__current_literal.clear()
        #

    def _append_literal_codepoint(self, codepoint: int) -> None:
        """Append the codepoint character to the current literal

        :param codepoint: the codepoint of the character to append
        """
        self.__current_literal.append(chr(codepoint))

    def _feed_characters(self) -> None:
        """Internal dispatcher method
        calling the appropriate processing methods
        """
        codepoint = self._pop_next_codepoint()
        if codepoint in self.quotes:
            self._process_quoted(codepoint)
        elif self.__subscript_start_position:
            self._process_in_subscript(codepoint)
        else:
            self._process_normal(codepoint)
        #

    def _pop_next_codepoint(self) -> int:
        """Pop the next character from the queue
        and increment the character position.

        :returns: the codepoint of the character
        """
        character = self.__query_characters.popleft()
        self.__character_position += 1
        return ord(character)

    def _process_quoted(self, quote_codepoint: int) -> None:
        """Process an opening quote and everything
        up to and including the expected closing quote.

        :param quote_codepoint: the quote codepoint to process
        :raises: UnclosedQouteException if the end of the
            query was encountered
        """
        self._add_current_literal()
        quote_started_at = self.__character_position
        expected_closing_quote = quote_codepoint
        self._append_literal_codepoint(quote_codepoint)
        while True:
            try:
                codepoint = self._pop_next_codepoint()
            except IndexError as error:
                raise UnclosedQuoteError(
                    character_position=quote_started_at
                ) from error
            #
            self._append_literal_codepoint(codepoint)
            if codepoint == expected_closing_quote:
                break
            #
        #
        self._add_current_literal()

    def _process_in_subscript(self, codepoint: int) -> None:
        """Process characters in subscript,
        treating equal signs and separators as literal components.

        :param codepoint: the codepoint to process
        """
        if codepoint in (
            self.subscript_indicators_pair.close,
            *self.whitespace,
        ):
            self._add_current_literal()
            item_constructor: Type[ItemWithContent]
            if codepoint == self.subscript_indicators_pair.close:
                self.__subscript_start_position = 0
                item_constructor = SubscriptCloserItem
            else:
                item_constructor = WhitespaceItem
            #
            self.__items_queue.append(item_constructor(chr(codepoint)))
        else:
            self._append_literal_codepoint(codepoint)
            if codepoint == self.subscript_indicators_pair.open:
                logging.warning(
                    "Possible error: Found subscript opener character %s (%r)"
                    " at position #%r inside a subscript",
                    hex(codepoint),
                    chr(codepoint),
                    self.__character_position,
                )
            #
        #

    def _process_normal(self, codepoint: int) -> None:
        """Normal processing of a single character from the query:
        opening subscript indicator, equals sign, separator,
        whitespace or literal content

        :param codepoint: the codepoint to process
        """
        if codepoint in (
            self.subscript_indicators_pair.open,
            self.separator_codepoint,
            self.equals_sign,
            *self.whitespace,
        ):
            self._add_current_literal()
            item_constructor: Type[ItemWithContent]
            if codepoint == self.subscript_indicators_pair.open:
                item_constructor = SubscriptOpenerItem
                self.__subscript_start_position = self.__character_position
            elif codepoint == self.equals_sign:
                item_constructor = AssignmentItem
            elif codepoint == self.separator_codepoint:
                item_constructor = SeparatorItem
            else:
                # implicit "codepoint in self.whitespace"
                item_constructor = WhitespaceItem
            #
            self.__items_queue.append(item_constructor(chr(codepoint)))
        else:
            self._append_literal_codepoint(codepoint)
        #

    def _reset(self) -> None:
        """Reset the itemizer"""
        self.__character_position = 0
        self.__current_literal.clear()
        self.__items_queue.clear()
        self.__query_characters.clear()
        self.__subscript_start_position = 0

    def itemize(self, original_query: str) -> Iterator[BaseItem]:
        """Yield items from a given query string

        :param original_query: a query string to be itemized
        """
        self._reset()
        self.__query_characters.extend(original_query)
        yield QueryStartItem()
        while self.__query_characters:
            self._feed_characters()
            while self.__items_queue:
                yield self.__items_queue.popleft()
            #
        #
        if self.__subscript_start_position:
            raise UnclosedSubscriptError(
                character_position=self.__subscript_start_position
            )
        #
        self._add_current_literal()
        while self.__items_queue:
            yield self.__items_queue.popleft()
        #
        yield QueryEndItem()


@dataclasses.dataclass(frozen=True)
class ParsedQuery:

    """Result of a query"""

    path: TraversalPath
    replacement: Optional[Tree]

    def __bool__(self) -> bool:
        """Returs False only if both the path is empty
        and the replacement is None
        """
        return bool(self.path) or self.replacement is not None

    def apply_to(self, original: Tree) -> Tree:
        """Apply the parsed query to the provided tree
        and return the result (a distinct Tree instance)
        """
        if self.replacement is None:
            return Tree(original.get_branch_clone(self.path))
        #
        new_tree = original.clone()
        new_tree.graft(self.path, self.replacement.root)
        return new_tree


class Parser:

    """Query parser"""

    phase_before_start = "before start"
    phase_started = "started"
    phase_address = "address"
    phase_replacement = "replacement"
    phase_ended = "ended"

    def __init__(self) -> None:
        """Initialize the itemizer internal state"""
        self.__phase = self.phase_before_start
        self.__collected_literals: List[LiteralItem] = []
        self.__collected_components: List[ScalarType] = []
        self.__replacement_fragments: List[str] = []

    @property
    def phase(self) -> str:
        """Property: current phase

        :returns: the current parser phase
        """
        return self.__phase

    def _append_path_component(self) -> None:
        """Add a path component from the collected literals

        Join adjacent literals, but yaml-safe-load them
        separately to enable mixing different quotes.
        Allow other literals than strings
        only if exactly one literal was collected.
        """
        if not self.__collected_literals:
            return
        #
        if len(self.__collected_literals) > 1:
            component = "".join(
                yaml.safe_load(item.content)
                for item in self.__collected_literals
            )
        else:
            component = yaml.safe_load(self.__collected_literals[0].content)
        #
        if isinstance(component, SCALAR_TYPES):
            self.__collected_components.append(component)
        else:
            raise TypeError(f"{component!r} is not a valid path component")
        #
        self.__collected_literals.clear()

    def _feed(self, item: BaseItem, position: int = 0) -> None:
        """Feed one item to the parser.
        Internal dispatcher method calling the appropriate handler
        according to the current phase.

        :param item: the item to be processed
        :param position: the item position (starting at 1)
        :raises: IllegalStateError if there is no handler defined
            for the current phase
        """
        for phase, method in (
            (self.phase_before_start, self._process_before_start),
            (self.phase_started, self._process_started),
            (self.phase_address, self._process_address),
            (self.phase_replacement, self._process_replacement),
            (self.phase_ended, self._process_ended),
        ):
            if phase == self.phase:
                item_handler = method
                break
            #
        else:
            raise IllegalStateError(
                "Item handler for this phase not implemented",
                offending_item=item,
                position=position,
                parser_phase=self.phase,
            )
        #
        item_handler(item, position=position)

    def _process_before_start(self, item: BaseItem, position: int = 0) -> None:
        """Process items in the "before_start" phase
        (Initial phase before processing the first item,
         which must unconditionally be a QueryStartItem).

        :param item: the item to be processed
        :param position: the item position (starting at 1)
        :raises: IllegalStateError if another item than a
            QueryStartItem is encountered
        """
        if isinstance(item, QueryStartItem):
            self.__phase = self.phase_started
            return
        #
        raise IllegalStateError(
            "Expected QueryStartItem()",
            offending_item=item,
            position=position,
            parser_phase=self.phase,
        )

    def _process_started(self, item: BaseItem, position: int = 0) -> None:
        """Process items in the "started" phase
        (Phase after the start item and before the
         unconditionally required initial SeparatorItem).
        Ignore whitespace, it is allowed and ignored here.
        Switch to the next phase when the SeparatorItem was found.

        :param item: the item to be processed
        :param position: the item position (starting at 1)
        :raises: IllegalStateError if neither a WhitespaceItem
            nor a SeparatorItem is encountered
        """
        if isinstance(item, WhitespaceItem):
            return
        #
        if isinstance(item, SeparatorItem):
            self.__phase = self.phase_address
            return
        #
        raise IllegalStateError(
            "The query must always start with a separator item",
            offending_item=item,
            position=position,
            parser_phase=self.phase,
        )

    def _process_address(self, item: BaseItem, position: int = 0) -> None:
        """Process items in the "address" phase
        (Phase after the initial SeparatorItem until
         either the first AssignmentItem or the final
         QueryEndItem is encountered).
        Collect literals for the address components.
        Switch phases if one of the aforementioned items was found.

        :param item: the item to be processed
        :param position: the item position (starting at 1)
        :raises: IllegalStateError if a QueryStartItem
            is encountered
        """
        if isinstance(item, QueryStartItem):
            raise IllegalStateError(
                "QueryStartItem() not allowed here",
                offending_item=item,
                position=position,
                parser_phase=self.phase,
            )
        #
        if isinstance(item, LiteralItem):
            self.__collected_literals.append(item)
            return
        #
        if isinstance(
            item,
            (
                AssignmentItem,
                QueryEndItem,
                SeparatorItem,
                SubscriptCloserItem,
                SubscriptOpenerItem,
            ),
        ):
            self._append_path_component()
        #
        if isinstance(item, AssignmentItem):
            self.__phase = self.phase_replacement
        elif isinstance(item, QueryEndItem):
            self.__phase = self.phase_ended
        #

    def _process_replacement(self, item: BaseItem, position: int = 0) -> None:
        """Process items in the "replacement" phase
        (Phase after the first AssignmentItem is encountered).
        Collect all Items having content for the replacement components.
        Switch phases if a QueryEndItem was found.

        :param item: the item to be processed
        :param position: the item position (starting at 1)
        :raises: IllegalStateError if a QueryStartItem()
            is encountered
        """
        if isinstance(item, QueryStartItem):
            raise IllegalStateError(
                "QueryStartItem() not allowed here",
                offending_item=item,
                position=position,
                parser_phase=self.phase,
            )
        #
        if isinstance(item, QueryEndItem):
            if not self.__replacement_fragments:
                self.__replacement_fragments.append("")
            #
            self.__phase = self.phase_ended
        elif isinstance(item, ItemWithContent):
            self.__replacement_fragments.append(item.content)
        #

    def _process_ended(self, item: BaseItem, position: int = 0) -> None:
        """Process items in the "ended" phase
        (Phase after the final QueryEndItem is encountered).
        No items allowed anymore in this phase.

        :param item: the item to be processed
        :param position: the item position (starting at 1)
        :raises: IllegalStateError unconditionally
        """
        raise IllegalStateError(
            "No items allowed anymore",
            offending_item=item,
            position=position,
            parser_phase=self.phase,
        )

    def _reset(self) -> None:
        """Reset the parser"""
        self.__phase = self.phase_before_start
        self.__collected_literals.clear()
        self.__collected_components.clear()
        self.__replacement_fragments.clear()

    def parse_query(
        self,
        original_query: str,
        separator_codepoint: Optional[int] = None,
        subscript_indicators_pair: Optional[commons.CharactersPair] = None,
    ) -> ParsedQuery:
        """Parse the given query and return a data structure address
        (access.ExtractingPath or access.ReplacingPath)

        :param original_query: a query string to be parsed
        :param separator_codepoint: (optionally) the codepoint
            of the separator
        :param subscript_indicators_pair: (optionally) the subscript
            indicators pair
        :returns: a tumple containg a TraversalPath instance and
            a Tree instance or None as replacement
        """
        self._reset()
        itemizer = Itemizer(
            separator_codepoint=separator_codepoint,
            subscript_indicators_pair=subscript_indicators_pair,
        )
        logging.debug("Found items in query:")
        for index, item in enumerate(itemizer.itemize(original_query)):
            position = index + 1
            logging.debug("%4d) %r", position, item)
            self._feed(item, position=position)
        #
        logging.debug("Path components: %r", self.__collected_components)
        path = TraversalPath(*self.__collected_components)
        replacement: Optional[Tree] = None
        if self.__replacement_fragments:
            logging.debug(
                "Replacement fragments: %r",
                self.__replacement_fragments,
            )
            serialized_replacement = "".join(
                self.__replacement_fragments
            ).strip()
            if not serialized_replacement:
                logging.warning("Assuming empty replacement value → None")
            #
            replacement = Tree.from_yaml(serialized_replacement)
        #
        logging.debug("Replacement: %r", replacement)
        return ParsedQuery(path, replacement)


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:
